/*
   Copyright 2005 Julian Robichaux

   Licensed under the Academic Free License version 2.1 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://opensource.org/licenses/afl-2.1.php

   Except as expressly stated in the License, software distributed 
   under the License is distributed on an "AS IS" BASIS, and WITHOUT 
   WARRANTY, either express or implied. See the License for the specific 
   language governing permissions and limitations under the License.

 */

package com.nsftools.jurst;


import java.io.*;
import java.net.*;
import java.util.*;
import javax.xml.parsers.*;
import org.w3c.dom.*;
import org.xml.sax.*;
import org.apache.soap.util.xml.*;
import org.apache.soap.*;
import org.apache.soap.rpc.*;
import org.apache.soap.encoding.*;
import org.apache.soap.transport.*;
import org.apache.soap.transport.http.*;
import javax.mail.MessagingException;

/**
 * The DebugCall class is an extension of the Apache SOAP RPC Call class, that
 * provides a little extra functionality that is useful for debugging.
 * <p>
 * I was really reluctant to write this class, but I did because I wanted an easy
 * way to get the raw XML of the SOAP response that was received, and there
 * didn't seem to be any way without using a sniffer or a proxy of some sort.
 * <p>
 * I knew that the raw response had to be handled by the Call class somewhere,
 * so I extended that class with this one and modified the <code>invoke</code> 
 * method so it stored the XML response internally instead of discarding it. Then
 * I added the <code>getRawSOAPMessage</code> and <code>getRawResponse</code>
 * methods, to provide easy access to the XML that was used for the outgoing SOAP
 * message and the incoming SOAP response for a given Call.
 * <p>
 * You should be able to use this DebugCall class in place of any Call classes you're
 * using, if you want to try it out.
 * <p>
 * Apologies to the authors of Apache SOAP for copying and then bastardizing the
 * code for the <code>invoke</code> method. But then again, the wonderful thing
 * about open-source software is that you can do that sort of thing, and add your
 * own functionality as needed.
 *
 * @author Julian Robichaux
 * @version 0.9.2
 */
 
public class DebugCall extends org.apache.soap.rpc.Call {
	private String xmlresponse = "";
	
	public DebugCall ()
	{
		this(null, null, null, null, null);
	}
	
	public DebugCall (String targetObjectURI, String methodName, Vector params,
		Header header, String encodingStyleURI)
	{
		this(targetObjectURI, methodName, params, header, encodingStyleURI,
			new SOAPContext());
	}
	
	public DebugCall (String targetObjectURI, String methodName, Vector params,
		Header header, String encodingStyleURI, SOAPContext ctx)
	{
		super(targetObjectURI, methodName, params, header, encodingStyleURI, ctx);
	}
	
	
	/**
	 * Get the XML text of the raw SOAP message that would be generated
	 * by this call, based on the parameters and information you've passed
	 * so far. If you've just called the invoke method, this should return
	 * the same XML SOAP message that was just sent.
	 * 
	 * @return	the raw XML of a SOAP message, as a String
	 */
	public String getRawSOAPMessage ()
	{
		try {
			java.io.StringWriter sw = new java.io.StringWriter();
			Envelope env = buildEnvelope();
			env.marshall(sw, getSOAPMappingRegistry(), getSOAPContext());
			return sw.toString();
		} catch (Exception e) {
			return e.getMessage();
		}
	}
	
	
	/**
	 * If the invoke method has already been called, this will
	 * return the SOAP response that was returned, as raw XML text.
	 * It the invoke method hasn't yet been called, you will get
	 * an empty String. If the invoke method returned an Exception,
	 * you will either get an empty String (if the error happened
	 * before or during transmission), or the raw response (if the 
	 * error occurred while trying to parse the response).
	 * 
	 * @return	the raw XML of a SOAP response, as a String
	 */
	public String getRawResponse ()
	{
		return xmlresponse;
	}
	
	
	/**
	 * This does the same thing as the invoke method in the parent
	 * <code>Call</code> class does, with some additional internal
	 * code that captures and stores the XML text of the SOAP response,
	 * if possible.
	 */
	public Response invoke (URL url, String SOAPActionURI) 
		throws SOAPException
	{
		xmlresponse = "";
		
		if (SOAPActionURI == null) {
			SOAPActionURI = "";
		}

		// if the smr hasn't been created yet, do it now
		SOAPMappingRegistry smr = getSOAPMappingRegistry();
		if (smr == null) {
			smr = new SOAPMappingRegistry();
		}

		try {
			// Build an envelope containing the call.
			Envelope callEnv = buildEnvelope();

			// Construct default HTTP transport if not specified.
			SOAPTransport st = getSOAPTransport();
			if (st == null)
				st = new SOAPHTTPConnection();

			// set the timeout
			int to = getTimeout();
			if (to != 0 && st instanceof SOAPHTTPConnection) 
				((SOAPHTTPConnection)st).setTimeout(to);

			// Post the call envelope.
			st.send(url, SOAPActionURI, null, callEnv, smr, ctx);

			// Get the response context.
			SOAPContext respCtx = st.getResponseSOAPContext();

			// Pre-read the response root part as a String to be able to
			// log it in an exception if parsing fails.
			//String payloadStr = getEnvelopeString(st);
			xmlresponse = getEnvelopeString(st);
			String payloadStr = xmlresponse;

			// Parse the incoming response stream.
			DocumentBuilder xdb = XMLParserUtils.getXMLDocBuilder();
			Document respDoc =
				xdb.parse(new InputSource(new StringReader(payloadStr)));
			Element payload = null;

			if (respDoc != null) {
				payload = respDoc.getDocumentElement();
			} else {
				throw new SOAPException (Constants.FAULT_CODE_CLIENT,
					"Parsing error, response was:\n" + payloadStr);
			}

			// Unmarshall the response envelope.
			Envelope respEnv = Envelope.unmarshall(payload, respCtx);

			// Extract the response from the response envelope.
			Response resp = Response.extractFromEnvelope(respEnv, smr, respCtx);

			// Reset the targetObjectURI in case it was changed by the server
			String fullTargetObjectURI = resp.getFullTargetObjectURI();

			if (fullTargetObjectURI != null) {
				setTargetObjectURI(fullTargetObjectURI);
			}

			return resp;
		}
		catch (MessagingException me) {
			throw new SOAPException(Constants.FAULT_CODE_CLIENT, me.getMessage(), me);
		} catch (IllegalArgumentException e) {
			throw new SOAPException(Constants.FAULT_CODE_CLIENT, e.getMessage(), e);
		} catch (SAXException e) {
			throw new SOAPException(Constants.FAULT_CODE_CLIENT,
									"Parsing error, response was:\n" +e.getMessage(), e);
		} catch (IOException e) {
			throw new SOAPException(Constants.FAULT_CODE_PROTOCOL, e.getMessage(), e);
		}
	}
	
}

