/*
   Copyright 2005 Julian Robichaux

   Licensed under the Academic Free License version 2.1 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://opensource.org/licenses/afl-2.1.php

   Except as expressly stated in the License, software distributed 
   under the License is distributed on an "AS IS" BASIS, and WITHOUT 
   WARRANTY, either express or implied. See the License for the specific 
   language governing permissions and limitations under the License.

 */

package com.nsftools.jurst;

import java.util.*;
import java.io.Serializable;
import javax.wsdl.Part;

/**
 * The <code>MethodInfo</code> class is a simple container
 * for information about a web service method, as retrieved from
 * a WSDL file. The information is primarily the information used
 * to build a proper SOAP call using the org.apache.soap classes.
 * <p>
 * This is used for passing around methods in the SoapHelper class.
 * The class also implements Serializable, in case you want to get
 * the MethodInfo from a WSDL file and store/retrieve it locally -- 
 * this is a lot more efficient than having to access and parse the
 * WSDL file every single time you want to call a method. 
 * 
 * @version 0.9.2
 * @author Julian Robichaux
 */
public class MethodInfo implements Serializable {
	/**
	 * serialVersionUID, required for implementing java.io.Serializable
	 */
	private static final long serialVersionUID = 8857646796509176519L;
	
	/*
	 * I may need to make some further changes to allow this
	 * class to be used in hash tables. For information, see:
	 * http://java.sun.com/developer/Books/effectivejava/Chapter3.pdf
	 */
	private String methodName;
	private String serviceName;
	private String portName;
	private String bindingName;
	private String namespace;
	private String soapAction;
	private String targetURL;
	private List paramList;
	private Map inputParts;
	private Map outputParts;
	
	/**
	 * @return Returns the bindingName.
	 */
	public String getBindingName() { return bindingName; }

	/**
	 * @param bindingName The bindingName to set.
	 */
	public void setBindingName(String bindingName) { this.bindingName = bindingName; }

	/**
	 * @return Returns the inputParts.
	 */
	public Map getInputParts() { return inputParts; }

	/**
	 * @param inputParts The inputParts to set.
	 */
	public void setInputParts(Map inputParts) { this.inputParts = inputParts; }

	/**
	 * @return Returns the methodName.
	 */
	public String getMethodName() { return methodName; }

	/**
	 * @param methodName The methodName to set.
	 */
	public void setMethodName(String methodName) { this.methodName = methodName; }

	/**
	 * @return Returns the namespace.
	 */
	public String getNamespace() { return namespace; }

	/**
	 * @param namespace The namespace to set.
	 */
	public void setNamespace(String namespace) { this.namespace = namespace; }

	/**
	 * @return Returns the outputParts.
	 */
	public Map getOutputParts() { return outputParts; }

	/**
	 * @param outputParts The outputParts to set.
	 */
	public void setOutputParts(Map outputParts) { this.outputParts = outputParts; }

	/**
	 * @return Returns the portName.
	 */
	public String getPortName() { return portName; }

	/**
	 * @param portName The portName to set.
	 */
	public void setPortName(String portName) { this.portName = portName; }

	/**
	 * @return Returns the serviceName.
	 */
	public String getServiceName() { return serviceName; }

	/**
	 * @param serviceName The serviceName to set.
	 */
	public void setServiceName(String serviceName) { this.serviceName = serviceName; }

	/**
	 * @return Returns the soapAction.
	 */
	public String getSoapAction() { return soapAction; }

	/**
	 * @param soapAction The soapAction to set.
	 */
	public void setSoapAction(String soapAction) { this.soapAction = soapAction; }

	/**
	 * @return Returns the targetURL.
	 */
	public String getTargetURL() { return targetURL; }

	/**
	 * @param targetURL The targetURL to set.
	 */
	public void setTargetURL(String targetURL) { this.targetURL = targetURL; }

	/**
	 * @param paramList The paramList to set.
	 */
	public void setParamList(List paramList) { this.paramList = paramList; }
	
	
	public String toString ()
	{
		StringBuffer sb = new StringBuffer();
		sb.append(methodName + "\n");
		sb.append("\tService: " + serviceName + "\n");
		sb.append("\tPort: " + portName + "\n");
		sb.append("\tBinding: " + bindingName + "\n");
		sb.append("\tNamespace: " + namespace + "\n");
		sb.append("\tSOAP Action: " + soapAction + "\n");
		sb.append("\tURL: " + targetURL + "\n");
		
		sb.append("\tInput Parameters: \n");
		sb.append(partsMapToString(inputParts));
		
		sb.append("\tOutput Parameters: \n");
		sb.append(partsMapToString(outputParts));
		
		return sb.toString();
	}
	
	private String partsMapToString (Map partsMap)
	{
		StringBuffer sb = new StringBuffer();
		try {
			int count = 1;
			for (Iterator it = partsMap.values().iterator(); it.hasNext(); ) {
				sb.append("\t\t" + count + ". ");
				Part part = (Part)it.next();
				sb.append(part.getName());
				sb.append(" (data type = " + part.getTypeName() + ")\n");
				count++;
			}
		} catch (Exception e) {
			sb.append("\t\tError: " + e + "\n");
		}
		
		return sb.toString();
	}
	
	public List getParamList ()
	{
		if (paramList != null)
			return paramList;
		if (inputParts == null)
			return null;
		
		ArrayList al = new ArrayList();
		for (Iterator it = inputParts.values().iterator(); it.hasNext(); ) {
			Part part = (Part)it.next();
			al.add(part.getName());
		}
		return al;
	}
	
}

