/*
   Copyright 2005 Julian Robichaux

   Licensed under the Academic Free License version 2.1 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://opensource.org/licenses/afl-2.1.php

   Except as expressly stated in the License, software distributed 
   under the License is distributed on an "AS IS" BASIS, and WITHOUT 
   WARRANTY, either express or implied. See the License for the specific 
   language governing permissions and limitations under the License.

 */

package com.nsftools.jurst;


import javax.wsdl.*;
import org.apache.soap.*;
import org.apache.soap.rpc.*;
import java.util.*;
import java.net.*;


/**
 * The <code>SoapHelper</code> class is an extension of the
 * <code>WsdlSoapWrapper</code> that simplifies calls to
 * SOAP-based web services even more than WsdlSoapWrapper
 * does.
 * <p>
 * You'll still have to have a little knowledge of the contents of the
 * WSDL file in many cases, but it's a lot easier than making all the calls
 * from scratch. To start with, you can use the SoapHelper class to give
 * you information about the available methods in a WSDL file, like this:
 * <p><blockquote><pre>
 * SoapHelper helper = new SoapHelper("http://localhost/EchoService?WSDL");
 * System.out.println(helper.dumpMethodList());
 * </pre></blockquote><p>
 * Here's an example of actually calling a web service:
 * <p><blockquote><pre>
 * SoapHelper sh = new SoapHelper("http://localhost/EchoService?WSDL");
 * ArrayList paramList = new ArrayList();
 * paramList.add("Does This Work?");
 * System.out.println("Response = " + sh.callSimpleMethodList("echoString", paramList));
 * </pre></blockquote><p>
 * Or, more concisely:
 * <p><blockquote><pre>
 * SoapHelper sh = new SoapHelper("http://localhost/EchoService?WSDL");
 * System.out.println("Response = " + sh.callSimpleMethod("echoString", "Does This Work?"));
 * </pre></blockquote><p>
 * If you need to pass or return a complex data type, you can do
 * something like this:
 * <p><blockquote><pre>
 * SoapHelper sh = new SoapHelper(wsdl);
 * // you'll need to have already created the BookInfo class
 * // that matches up to the web service's BookInfo data type
 * sh.addComplexDatatype("BookInfo", BookInfo.class);
 * Object obj = sh.callSimpleMethod("getBookInfo", "Cryptomonicon");
 * // the response will be a String if there was a Fault; otherwise
 * // it's supposed to be an instance of BookInfo
 * if (obj instanceof String) {
 * 	System.out.println("String response = " + obj);
 * } else if (obj instanceof BookInfo) {
 * 	BookInfo bi = (BookInfo)obj;
 * 	System.out.println("BooksInfo response " + bi);
 * } else {
 * 	System.out.println("Response was a " + obj.getClass().getName());
 * }
 * </pre></blockquote><p>
 * You can use the callMethod routine from the parent 
 * WsdlSoapWrapper class if you need a little more control over 
 * the parameters you send and the response that you get back,
 * but I've found that in most cases I can get away with one of the
 * callSimpleMethod calls. See the comments in the
 * WsdlSoapWrapper class for more details and information.
 * <p>
 * AUTHOR'S NOTE:<br>
 * Okay, now this whole overloaded callSimpleMethod 
 * business is UG-LY. But Java didn't introduce the concept
 * of varargs until Java 1.5, and lots of people are still on
 * older versions of Java right now. So we've got 11 differerent
 * versions of callSimpleMethod -- with no params, with one
 * param, with two params, etc. All that, just to make it easier
 * for the user to call a web service. So instead of:
 * <p><blockquote><pre>
 *	ArrayList paramList = new ArrayList();
 *	paramList.add("whatever");
 *	Object obj = sh.callSimpleMethodList("method1", paramList);
 * </pre></blockquote><p>
 * you can just call:
 * <p><blockquote><pre>
 * 	Object obj = sh.callSimpleMethod("method1", "whatever");
 * </pre></blockquote><p>
 * Granted, it's only saving 2 lines of code, but sometimes
 * that's what makes people happy.	  		
 * 
 * @author Julian Robichaux
 * @version 0.9.2
 */

public class SoapHelper extends WsdlSoapWrapper {
	
	/**
	 * did a fault get returned when we called callSimpleMethodList
	 * or callSimpleMethod?
	 */
	private boolean faulted = false;
	
	
	/**
	 * Create a new SoapHelper object that will reference
	 * a WSDL file at the given location (can be an http://
	 * or a file:/// path)
	 *
	 * @param wsdlURI	a String giving the path to the
	 * 								WSDL file we'll be referencing
	 * 								(can be "http://..." or "file:///...")
	 */
	public SoapHelper (String wsdlURI)
	{
		super(wsdlURI);
	}
	
	
	/**
	 * Call the method with the given name, using the list
	 * of parameters that are passed, and return the Java 
	 * object that is generated by the Response (or a String
	 * containing Fault information in the case of a fault).
	 * <p>
	 * The difference between this and <code>callMethod</code>
	 * is with this you pass a List of Objects that should be used as
	 * parameters, and the Objects in that List are included <i>in the
	 * order specified</i> by the parameter list in the WSDL file 
	 * (whereas with callMethod you match the parameter name
	 * to the object for each Parameter object in the Vector). Also,
	 * callMethod returns the actual Response object generated by
	 * the SOAP call, while this method returns the object contained
	 * within the Response object (or a String containing Fault data).
	 * 
	 * @param methodName	the method that we're calling
	 * 										(if the WSDL file has overloaded
	 * 										the method, the first one found
	 * 										will be called)
	 * @param paramValues	List of Objects, one for each parameter 
	 * 									you want to pass to the call (or null if no 
	 * 									parameters are desired)
	 * @return		the Java Object contained within the Response from
	 * 					this call, or a String containing Fault data if there was
	 * 					an error on the web service side (normal Java
	 * 					Exceptions will be thrown in the case of runtime
	 * 					errors on the client or connectivity side)
	 */
	public Object callSimpleMethodList (String methodName, List paramValues) 
		throws MalformedURLException, SOAPException, WSDLException
	{
		faulted = false;
		MethodInfo mi = getMethod(methodName, isCaseSensitive());
		return callSimpleMethodList(mi, paramValues);
	}
	
	
	/**
	 * Same as callSimpleMethodList(methodName, paramValues), except
	 * you pass your own MethodInfo object instead of a method name.
	 * See documentation of the other version of this method for more information.
	 */
	public Object callSimpleMethodList (MethodInfo mi, List paramValues) 
		throws MalformedURLException, SOAPException, WSDLException
	{
		faulted = false;
		Vector v = new Vector();
		List paramList = mi.getParamList();
		
		if (paramValues != null) {
			for (int i = 0; i < paramValues.size(); i++)
			{
				if (i < paramList.size()) {
					if (paramValues.get(i) != null) {
						String paramName = (String)paramList.get(i);
						Object obj = paramValues.get(i);
						v.add(new Parameter(paramName, obj.getClass(), obj, null));
						debugPrint("Added parameter " + paramName + ": " + obj);
					}
				}
			}
		}
		
		Response resp = callMethod(mi.getMethodName(), v);
		if (!resp.generatedFault())
		{
			Parameter result = resp.getReturnValue();
			return result.getValue();
		} else {
			faulted = true;
			org.apache.soap.Fault f = resp.getFault();
			return "Fault code " + f.getFaultCode() + ": " + f.getFaultString();
		}
		
	}
	
	
	/**
	 * Returns true if the last call to callSimpleMethod resulted in a
	 * Fault, false otherwise (note that this is NOT set by callMethod,
	 * only by callSimpleMethod).
	 */
	public boolean isFault ()
	{
		return faulted;
	}
	
	
	/**
	 * Overloaded version of callSimpleMethod that allows
	 * you to easily call a method that has no parameters.
	 * This method will end up calling callSimpleMethodList,
	 * so please see the documentation for that method for
	 * information about usage and return values.
	 */
	public Object callSimpleMethod (String methodName) 
		throws MalformedURLException, SOAPException, WSDLException
	{
		return callSimpleMethodList(methodName, null);
	}
	
	
	/**
	 * Overloaded version of callSimpleMethod that allows
	 * you to easily call a method that has one parameter.
	 * This method will end up calling callSimpleMethodList,
	 * so please see the documentation for that method for
	 * information about usage and return values.
	 * <p>
	 * The values that are passed can be any type of object 
	 * (String, Double, custom class, etc.) that is an appropriate 
	 * parameter for the web service method that is being called.
	 * You can also pass an Object array (String[], Double[], etc.)
	 * as a parameter, if appropriate.
	 * <p>
	 * If you wish to pass a primitive as a parameter, please use
	 * the corresponding Java Object instead (i.e. -- Integer
	 * instead of integer).
	 */
	public Object callSimpleMethod (String methodName, Object value1) 
		throws MalformedURLException, SOAPException, WSDLException
	{
		return internalCallSimpleMethod(methodName, 1, value1, 
			null, null, null, null, null, null, null, null, null);
	}
	
	
	/**
	 * Easily call a method that has two parameters.
	 * See callSimpleMethod(methodName, value1) for details.
	 */
	public Object callSimpleMethod (String methodName, Object value1, 
		Object value2) 
		throws MalformedURLException, SOAPException, WSDLException
	{
		return internalCallSimpleMethod(methodName, 2, value1, 
			value2, null, null, null, null, null, null, null, null);
	}
	
	
	/**
	 * Easily call a method that has three parameters.
	 * See callSimpleMethod(methodName, value1) for details.
	 */
	public Object callSimpleMethod (String methodName, Object value1, 
		Object value2, Object value3) 
		throws MalformedURLException, SOAPException, WSDLException
	{
		return internalCallSimpleMethod(methodName, 3, value1, 
			value2, value3, null, null, null, null, null, null, null);
	}
	
	
	/**
	 * Easily call a method that has four parameters.
	 * See callSimpleMethod(methodName, value1) for details.
	 */
	public Object callSimpleMethod (String methodName, Object value1, 
		Object value2, Object value3, Object value4) 
		throws MalformedURLException, SOAPException, WSDLException
	{
		return internalCallSimpleMethod(methodName, 4, value1, 
			value2, value3, value4, null, null, null, null, null, null);
	}
	
	
	/**
	 * Easily call a method that has five parameters.
	 * See callSimpleMethod(methodName, value1) for details.
	 */
	public Object callSimpleMethod (String methodName, Object value1, 
		Object value2, Object value3, Object value4, Object value5) 
		throws MalformedURLException, SOAPException, WSDLException
	{
		return internalCallSimpleMethod(methodName, 5, value1, 
			value2, value3, value4, value5, null, null, null, null, null);
	}
	
	
	/**
	 * Easily call a method that has six parameters.
	 * See callSimpleMethod(methodName, value1) for details.
	 */
	public Object callSimpleMethod (String methodName, Object value1, 
		Object value2, Object value3, Object value4, Object value5, 
		Object value6) 
		throws MalformedURLException, SOAPException, WSDLException
	{
		return internalCallSimpleMethod(methodName, 6, value1, 
			value2, value3, value4, value5, value6, null, null, null, null);
	}
	
	
	/**
	 * Easily call a method that has seven parameters.
	 * See callSimpleMethod(methodName, value1) for details.
	 */
	public Object callSimpleMethod (String methodName, Object value1, 
		Object value2, Object value3, Object value4, Object value5, 
		Object value6, Object value7) 
		throws MalformedURLException, SOAPException, WSDLException
	{
		return internalCallSimpleMethod(methodName, 7, value1, 
			value2, value3, value4, value5, value6, value7, null, null, null);
	}
	
	
	/**
	 * Easily call a method that has eight parameters.
	 * See callSimpleMethod(methodName, value1) for details.
	 */
	public Object callSimpleMethod (String methodName, Object value1, 
		Object value2, Object value3, Object value4, Object value5, 
		Object value6, Object value7, Object value8) 
		throws MalformedURLException, SOAPException, WSDLException
	{
		return internalCallSimpleMethod(methodName, 8, value1, 
			value2, value3, value4, value5, value6, value7, value8, null, null);
	}
	
	
	/**
	 * Easily call a method that has nine parameters.
	 * See callSimpleMethod(methodName, value1) for details.
	 */
	public Object callSimpleMethod (String methodName, Object value1, 
		Object value2, Object value3, Object value4, Object value5, 
		Object value6, Object value7, Object value8, Object value9) 
		throws MalformedURLException, SOAPException, WSDLException
	{
		return internalCallSimpleMethod(methodName, 9, value1, 
			value2, value3, value4, value5, value6, value7, value8, value9, null);
	}
	
	
	/**
	 * Easily call a method that has ten parameters.
	 * See callSimpleMethod(methodName, value1) for details.
	 */
	public Object callSimpleMethod (String methodName, Object value1, 
		Object value2, Object value3, Object value4, Object value5, 
		Object value6, Object value7, Object value8, Object value9, Object value10) 
		throws MalformedURLException, SOAPException, WSDLException
	{
		return internalCallSimpleMethod(methodName, 10, value1, 
			value2, value3, value4, value5, value6, value7, value8, value9, value10);
	}
	
	
	/**
	 * Here's what callSimpleMethod is actually doing
	 */
	private Object internalCallSimpleMethod (String methodName, int count, 
		Object value1, Object value2, Object value3, Object value4, Object value5, 
		Object value6, Object value7, Object value8, Object value9, Object value10) 
		throws MalformedURLException, SOAPException, WSDLException
	{
		ArrayList paramList = new ArrayList();
		if (count > 0) { paramList.add(value1); }
		if (count > 1) { paramList.add(value2); }
		if (count > 2) { paramList.add(value3); }
		if (count > 3) { paramList.add(value4); }
		if (count > 4) { paramList.add(value5); }
		if (count > 5) { paramList.add(value6); }
		if (count > 6) { paramList.add(value7); }
		if (count > 7) { paramList.add(value8); }
		if (count > 8) { paramList.add(value9); }
		if (count > 9) { paramList.add(value10); }
		return callSimpleMethodList(methodName, paramList);
	}
	
	
	/**
	 * Returns a String value containing information about all
	 * the methods in the WSDL file that was given when this 
	 * object was instantiated. This is really just a convenience,
	 * for times when you want to know how this class sees the
	 * WSDL contents.
	 *
	 * @return		a String with information about all the methods
	 * 					in this WSDL file
	 */
	public String dumpMethodList ()
	{
		StringBuffer sb = new StringBuffer();
		
		try {
			Vector methodList = getMethodList();
			
			for (int i = 0; i < methodList.size(); i++)
			{
				MethodInfo mi = (MethodInfo)methodList.get(i);
				sb.append(mi + "\n\n");
			}
		} catch (Exception e) {
			sb.append(e.getMessage());
		}
		
		return sb.toString();
	}
	
	
	/**
	 * Returns a String value containing information about the
	 * data types (standard or complex) in the WSDL file that 
	 * was given when this object was instantiated.
	 * <p>
	 * The type list is generated by looking at the input and output
	 * message definitions for each method, so it's possible that 
	 * there are complex data types that are defined in the WSDL file 
	 * that won't be listed here, because they aren't used as 
	 * parameters or return values. So, for example, a complex data
	 * type that is nested within another complex data type might
	 * not show up on this list. 
	 *
	 * @return		a String with information about the data types
	 * 					in this WSDL file that are used as parameters
	 * 					for request or response messages
	 */
	public String dumpTypeList ()
	{
		StringBuffer sb = new StringBuffer();
		
		try {
			Map typeMap = getTypes();
			
			if (typeMap == null) {
				sb.append("No type references found\n");
			} else {
				int count = 0;
				for (Iterator it = typeMap.keySet().iterator(); it.hasNext(); ) {
					String typeName = (String)it.next();
					sb.append(typeName + "\n");
					sb.append("\tNamespace: " + typeMap.get(typeName) + "\n\n");
					count++;
				}
			}
		} catch (Exception e) {
			sb.append(e.getMessage());
		}
		
		return sb.toString();
		
	}

}

