/*
   Copyright 2005 Julian Robichaux

   Licensed under the Academic Free License version 2.1 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://opensource.org/licenses/afl-2.1.php

   Except as expressly stated in the License, software distributed 
   under the License is distributed on an "AS IS" BASIS, and WITHOUT 
   WARRANTY, either express or implied. See the License for the specific 
   language governing permissions and limitations under the License.

 */

package com.nsftools.jurst;

import javax.wsdl.*;
import javax.wsdl.xml.*;
import javax.wsdl.factory.*;
import javax.wsdl.extensions.*;
import javax.wsdl.extensions.soap.*;
import org.apache.soap.*;
import org.apache.soap.rpc.*;
import org.apache.soap.encoding.*;
import org.apache.soap.encoding.soapenc.*;
import org.apache.soap.transport.http.*;
import org.apache.soap.util.xml.*;
import java.net.*;
import java.util.*;
import java.io.*;

/**
 * The <code>WsdlSoapWrapper</code> class provides a simple way
 * to call a web service using Apache SOAP and WSDL4J, without 
 * necessarily having to read the WSDL file yourself and figure out 
 * what the methods and namespaces and parameters should be.
 * <p>
 * Okay, you'll have to have a little knowledge of the contents of the
 * WSDL file in many cases, but it's a lot easier than making all the calls
 * from scratch.
 * <p>
 * Here's an example of actually calling a web service:
 * <p><blockquote><pre>
 * WsdlSoapWrapper wsw = new WsdlSoapWrapper("http://localhost/EchoService?WSDL");
 * Vector params = new Vector();
 * String p1 = "Hello";
 * params.add(new Parameter("ParamName", p1.getClass(), p1, null));
 * Response resp = callMethod("testMethod", params);
 * if (!resp.generatedFault()) {
 * 	Parameter result = resp.getReturnValue();
 * }
 * </pre></blockquote><p>
 * If you need to pass or return a complex data type, you can use
 * addComplexDatatype to register a Class to handle the type like this:
 * <p><blockquote><pre>
 * WsdlSoapWrapper wsw = new WsdlSoapWrapper("http://localhost/EchoService?WSDL");
 * wsw.addComplexDatatype("BookInfo", BookInfo.class);
 * </pre></blockquote><p>
 * See the comments in the SoapHelper class (which is a subclass
 * of WsdlSoapWrapper) for more examples. In fact, in most cases
 * you can make your SOAP calls with callSimpleMethod in
 * SoapWrapper much more easily than using callMethod here.
 *<p>
 * Other items of note:
 * <p><ul>
 * <li>the <code>caseSensitive</code> variable controls
 * whether the method names in the WSDL file are looked up in a
 * case-sensitive or non-case-sensitive manner (by default, this is
 * false, so it's NOT case-sensitive)</li>
 * <li>you can set the <code>timeout</code> variable to a 
 * number greater than zero to enforce a timeout value when calling
 * a web service (the timeout is in milliseconds)</li>
 * <li>You can also send internal debug messages to a PrintStream
 * (like System.err or a file) by setting <code>setDebugOut</code>.
 * This will show you, among other things, the raw XML SOAP messages
 * that are sent out as requests and the raw XML SOAP responses that
 * are received.</li>
 * </ul>
 * 
 * @author Julian Robichaux
 * @version 0.9.2
 */
public class WsdlSoapWrapper {
	
	private boolean caseSensitive = false;
	private int timeout = 0;
	
	private Vector methodList = null;
	private String wsdl;
	private SOAPMappingRegistry smr = null;
	private PrintStream debugStream = null;
	private SOAPHTTPConnection con = null;
	private HashMap typeMap = null;
	private String lastXmlResponse = "";
	private String basicAuthHeader = "";
	private String proxyAuthHeader = "";
	
	/**
	 * @return Returns the caseSensitive.
	 */
	public boolean isCaseSensitive() {
		return caseSensitive;
	}


	/**
	 * Determines whether or not the method names are
	 * case-sensitive. Default is false (NOT case-sensitive).
	 */
	public void setCaseSensitive(boolean caseSensitive) {
		this.caseSensitive = caseSensitive;
	}


	/**
	 * @return Returns the timeout.
	 */
	public int getTimeout() {
		return timeout;
	}


	/**
	 * Specifies a connection timeout value in milliseconds,
	 * if any. Default is 0 (no timeout).
	 */
	public void setTimeout(int timeout) {
		this.timeout = timeout;
	}


	/**
	 * Create a new WsdlSoapWrapper object that will reference
	 * a WSDL file at the given location (can be an http://
	 * or a file:/// path)
	 *
	 * @param wsdlURI	a String giving the path to the
	 * 								WSDL file we'll be referencing
	 * 								(can be "http://..." or "file:///...")
	 */
	public WsdlSoapWrapper (String wsdlURI)
	{
		wsdl = wsdlURI;
	}
	
	
	/**
	 * Set the PrintStream that you want debug messages
	 * to be sent to, if any (by default this is null). Common
	 * values are System.err or a PrintStream that is wrapped
	 * around a FileOutputStream.
	 * 
	 * @param stream		the PrintStream to write to
	 */
	public void setDebugOut (PrintStream stream)
	{
		debugStream = stream;
	}
	
	
	/**
	 * Returns the PrintStream that we're sending debug
	 * output to, if any (this is null by default).
	 *
	 * @return 		the current debug PrintStream
	 */
	public PrintStream getDebugOut ()
	{
		return debugStream;
	}
	
	
	/**
	 * If a debugStream has been set (by default it's null),
	 * try to print debug information to the Stream. Debugging
	 * messages are scattered throughout various methods
	 * in this class.
	 * <p>
	 * If you just want to see what it does, you can
	 * call setDebugOut(System.out) on the SoapHelper object
	 * you create and use it as you normally would.
	 */
	public void debugPrint (Object obj)
	{
		if (debugStream != null) {
			try {
				debugStream.println("DEBUG " + obj.toString());
			} catch (Exception whatever) {
			}
		}
	}
	
	
	/**
	 * If basic authentication is used to connect to the web service, enter
	 * the user name and password here.
	 */
	public void setBasicAuthentication (String username, String password)
	{
		if (con == null)
			con = new SOAPHTTPConnection();
		con.setUserName(username);
		con.setPassword(password);
		basicAuthHeader = encodeHeaderAuthentication(username, password);
	}
	
	
	/**
	 * If you must authenticate through a proxy server to connect to the 
	 * web service, enter the proxy server host name, port, username, and
	 * password here.
	 */
	public void setProxyAuthentication (String proxyServer, int port, 
		String username, String password)
	{
		if (con == null)
			con = new SOAPHTTPConnection();
		con.setProxyHost(proxyServer);
		con.setProxyPort(port);
		con.setProxyUserName(username);
		con.setProxyPassword(password);
		
		proxyAuthHeader = encodeHeaderAuthentication(username, password);
	}
	
	
	private String encodeHeaderAuthentication (String username, String password)
	{
		if (username == null) { username = ""; }
		if (password == null) { password = ""; }
		
		if (username.length() > 0) {
			try {
				return Base64.encode((username + ":" + password).getBytes("8859_1"));
			} catch (Exception e) {}
		}
		
		return "";
	}
	
	
	/**
	 * Get the last response that was received via callMethod as an XML String,
	 * if possible. This can be helpful in debugging errors.
	 */
	public String getLastXmlResponse () 
	{
		return lastXmlResponse;
	}
	
	
	/**
	 * Call the method with the given name, using the list
	 * of parameters that are passed, and return the Response
	 * object that is generated.
	 * 
	 * @param methodName	the method that we're calling
	 * 										(if the WSDL file has overloaded
	 * 										the method, the first one found
	 * 										will be called)
	 * @param params		Vector of org.apache.soap.rpc.Parameter
	 * 								objects, one for each parameter you want
	 * 								to pass to the call (or null if no parameters
	 * 								are desired)
	 * @return		the Response object generated by this call, which
	 * 					will have the generatedFault() = true if there was
	 * 					an error on the web service side (normal Java
	 * 					Exceptions will be thrown in the case of runtime
	 * 					errors on the client or connectivity side)
	 */
	public Response callMethod (String methodName, Vector params) 
		throws MalformedURLException, SOAPException, WSDLException
	{
		MethodInfo mi = getMethod(methodName, caseSensitive);
		return callMethod(mi, params);
	}
	
	
	/**
	 * Same as callMethod(methodName, params), except you pass a
	 * MethodInfo object instead of a method name. See documentation
	 * for that version of this method for more information on usage.
	 * <p>
	 * This is good if you want to save the effort of accessing and parsing 
	 * a WSDL file (which is necessary if you just pass a method name) because
	 * you've stored away the MethodInfo information somewhere, or if
	 * you'd rather pass your own hardcoded MethodInfo values, rather
	 * than the ones that are automatically generated by reading the
	 * WSDL file.
	 */
	public Response callMethod (MethodInfo mi, Vector params) 
		throws MalformedURLException, SOAPException, WSDLException
	{
		debugPrint("Starting callMethod: " + (new Date()));
		
		//Call call = new Call();
		DebugCall call = new DebugCall();
		lastXmlResponse = "";
		
		if (smr != null) {
			call.setSOAPMappingRegistry(smr);
		}
		if (timeout > 0) {
			call.setTimeout(timeout);
		}
		if (con == null) {
			con = new SOAPHTTPConnection();
		}
		
		call.setSOAPTransport(con);
		call.setTargetObjectURI(mi.getNamespace());
		call.setMethodName(mi.getMethodName());
		call.setEncodingStyleURI(org.apache.soap.Constants.NS_URI_SOAP_ENC);
		call.setParams(params);
		
		Response resp = null;
		try {
			resp = call.invoke(new URL(mi.getTargetURL()), mi.getSoapAction());
		} catch (SOAPException sex) {
			throw sex;
		} finally {
			// even if there's an error, try to get the message and response
			lastXmlResponse = call.getRawResponse();
			
			// debug output
			if (debugStream != null) {
				debugPrint("Call =\n" + call.getRawSOAPMessage());
				debugPrint("Response =\n" + call.getRawResponse());
			}
		}
		
		debugPrint("Ending callMethod: " + (new Date()));
		return resp;
	}
	
	
	/**
	 * Attempt to get the WSDL file at the given URI as a Definition,
	 * using any of the authentication or proxy parameters that have
	 * been specified so far. Normally, the URI will be something like
	 * http://myserver/file.wsdl or file:///c:/file.wsdl
	 * <p>
	 * I'm not entirely sure this will work if the URI is specified as
	 * an https:// path. It's possible, but you would at least need to
	 * have JSSE included and configured properly. For more information
	 * on JSSE, see:<br>
	 * http://java.sun.com/j2se/1.4.2/docs/guide/security/jsse/JSSERefGuide.html
	 * 
	 * @param wsdlURI	a String giving the path to the
	 * 						WSDL file we'll be referencing
	 * 						(can be "http://..." or "file:///...")
	 * @return	a WSDL4J Definition based on the WSDL file
	 */
	public Definition getWsdlDef (String wsdlURI)
		throws WSDLException, IOException, MalformedURLException
	{
		Definition def = null;
		
		debugPrint("Starting getWsdlDef: " + (new Date()));
		
		//WSDLReaderImpl reader = new WSDLReaderImpl();
		WSDLFactory wsdlFactory = WSDLFactory.newInstance();
		WSDLReader reader = wsdlFactory.newWSDLReader();
		reader.setFeature("javax.wsdl.verbose", false);
		
		String protocol = "";
		if (wsdlURI.indexOf(":") >= 0)
			protocol = wsdlURI.substring(0, wsdlURI.indexOf(":")).toLowerCase();
		
		if (protocol.startsWith("http")) {
			URL theUrl;
			
			String proxyHost = "";
			if (con != null)
				proxyHost = (con.getProxyHost() == null) ? "" : con.getProxyHost();
			
			if (proxyHost.length() > 0) {
				theUrl = new URL(protocol, proxyHost, con.getProxyPort(), wsdlURI);
			}  else  {
				theUrl = new URL(wsdlURI);
			}
			
			URLConnection con = theUrl.openConnection();
			
			if (basicAuthHeader.length() > 0)
				con.setRequestProperty("Authorization", basicAuthHeader);
			if (proxyAuthHeader.length() > 0)
				con.setRequestProperty("Proxy-Authorization", proxyAuthHeader);
			
			BufferedInputStream in = new BufferedInputStream(con.getInputStream());
			org.xml.sax.InputSource input = new org.xml.sax.InputSource(in);
			input.setEncoding(con.getContentEncoding());
			
			def = reader.readWSDL(null, input);
			in.close();
			
		} else {
			def = reader.readWSDL(null, wsdlURI);
		}

		debugPrint("Ending getWsdlDef: " + (new Date()));
		return def;

	}
	
	/**
	 * For the WSDL file at the given location, return a Vector
	 * containing a MethodInfo object for each method in the
	 * WSDL file.
	 *
	 * @param wsdlURI	a String giving the path to the
	 * 								WSDL file we'll be referencing
	 * 								(can be "http://..." or "file:///...")
	 * @return		a Vector of MethodInfo objects
	 */
	public Vector getAllMethods (String wsdlURI)
		throws WSDLException
	{
		Vector v = new Vector();
		
		/*
		WSDLReaderImpl reader = new WSDLReaderImpl();
		reader.setFeature("javax.wsdl.verbose", false);
		Definition def = reader.readWSDL(null, wsdlURI);
		*/
		Definition def = null;
		try {
			def = getWsdlDef(wsdlURI);
		} catch (MalformedURLException mue) {
			throw new WSDLException(WSDLException.OTHER_ERROR, mue.getMessage());
		} catch (IOException ioe) {
			throw new WSDLException(WSDLException.OTHER_ERROR, ioe.getMessage());
		}
		
		Map services = def.getServices();
		for (Iterator it1 = services.values().iterator(); it1.hasNext(); )
		{
			Service svc = (Service)it1.next();
			debugPrint("Found service: " + svc.getQName().getLocalPart());
			
			Map ports = svc.getPorts();
			for (Iterator it2 = ports.values().iterator(); it2.hasNext(); )
			{
				Port port = (Port)it2.next();
				debugPrint("Found port: " + port.getName());
				
				Binding binding = port.getBinding();
				List bindOps = binding.getBindingOperations();
				for (int i = 0; i < bindOps.size(); i++)
				{
					BindingOperation bindOp = (BindingOperation)bindOps.get(i);
					
					debugPrint("Found method: " + bindOp.getName());
					MethodInfo mi = new MethodInfo();
					mi.setMethodName(bindOp.getName());
					mi.setServiceName(svc.getQName().getLocalPart());
					mi.setPortName(port.getName());
					mi.setBindingName(binding.getQName().getLocalPart());
					mi.setParamList(bindOp.getOperation().getParameterOrdering());
					mi.setInputParts(bindOp.getOperation().getInput().getMessage().getParts());
					mi.setOutputParts (bindOp.getOperation().getOutput().getMessage().getParts());
					mi.setTargetURL(getPortSoapURL(port));
					mi.setSoapAction(getMethodSoapAction(bindOp));
					mi.setNamespace(getMethodInputNamespace(bindOp));
					if ((mi.getNamespace() == null) || (mi.getNamespace().equals(""))) {
						mi.setNamespace(binding.getQName().getNamespaceURI());
					}
					
					v.add(mi);
					addTypes(bindOp.getOperation());
				}
			}
			
		}
		
		return v;
	}
	
	
	// another possible way to parse and deal with types here:
	// http://forum.java.sun.com/thread.jspa?threadID=467761&tstart=0
	private void addTypes (Operation op)
	{
		addTypes(op.getInput().getMessage().getParts());
		addTypes(op.getOutput().getMessage().getParts());
	}
	
	
	private void addTypes (Map partsMap)
	{
		if (typeMap == null)
			typeMap = new HashMap();
		
		for (Iterator it = partsMap.values().iterator(); it.hasNext(); ) {
			Part part = (Part)it.next();
			if (part.getTypeName() != null)
				typeMap.put(part.getTypeName().getLocalPart(), part.getTypeName().getNamespaceURI());
		}
	}
	
	
	/**
	 * Return a Map of data types (standard and complex) that
	 * are used as Input or Output parameters in this WSDL file.
	 * The key for each Map entry is the name of the data type,
	 * and the value for each entry is that type's namespace.
	 *  
	 * @return a Map of data type names and their namespaces
	 * 
	 * @throws WSDLException
	 */
	public Map getTypes ()
		throws WSDLException
	{
		if (methodList == null)
			methodList = getAllMethods(wsdl);
		return typeMap;
	}
	
	
	/**
	 * Look for a method with the given methodName in the given
	 * WSDL file, and return a MethodInfo object with information
	 * about it. A WSDLException will be thrown if the method is
	 * not found.
	 *
	 * @param methodName	a String giving the name of
	 * 										the method we're looking for
	 * 										(the match IS case-sensitive)
	 * @return		a Vector of MethodInfo objects
	 */
	public MethodInfo getMethod (String methodName)
		throws WSDLException
	{
		return getMethod(methodName, true);
	}
	
	
	/**
	 * Look for a method with the given methodName in the given
	 * WSDL file, and return a MethodInfo object with information
	 * about it. A WSDLException will be thrown if the method is
	 * not found.
	 *
	 * @param methodName	a String giving the name of
	 * 										the method we're looking for
	 * @param caseSensitive	whether or not the search for
	 * 										the methodName is case-sensitive
	 * 										(true) or not (false)
	 * @return		a Vector of MethodInfo objects
	 */
	public MethodInfo getMethod (String methodName, boolean caseSensitive) 
		throws WSDLException
	{
		if (methodList == null)
			methodList = getAllMethods(wsdl);
		
		for (int i = 0; i < methodList.size(); i++)
		{
			MethodInfo mi = (MethodInfo)methodList.get(i);
			if (mi.getMethodName().equals(methodName)) {
				return mi;
			} else if ((!caseSensitive) && 
				(mi.getMethodName().toUpperCase().equals(methodName.toUpperCase()))) {
				return mi;
			}
		}
		
		throw new WSDLException("Method not found", 
				"Method " + methodName + " not found at " + wsdl);
	}
	
	
	/**
	 * Returns a Vector of MethodInfo objects, one for each method
	 * in the WSDL file that was given when this object was instantiated.
	 *
	 * @return		a Vector of MethodInfo objects
	 */
	public Vector getMethodList () 
		throws WSDLException
	{
			if (methodList == null)
				methodList = getAllMethods(wsdl);
			return methodList;
	}
	
	
	/**
	 * Given a BindingOperation in a WSDL file, try to get the
	 * SOAP Action that is specified for that operation, if
	 * one has been specified.
	 *
	 * @param bindOp		the BindingOperation we're looking at
	 * @return		the SOAP Action for this BindingOperation, if one
	 * 					could be found (null or "" if not)
	 */
	public String getMethodSoapAction (BindingOperation bindOp)
	{
		List extEls = bindOp.getExtensibilityElements();
		for (int j = 0; j < extEls.size(); j++)
		{
			ExtensibilityElement extEl = (ExtensibilityElement)extEls.get(j);
			if (extEl instanceof SOAPOperation)
			{
				return ((SOAPOperation)extEl).getSoapActionURI();
			}
		}
		
		return null;
	}
	
	
	/**
	 * Given a BindingOperation in a WSDL file, try to get the
	 * namespace that is specified for the input message, if
	 * one has been specified.
	 *
	 * @param bindOp		the BindingOperation we're looking at
	 * @return		the namespace for this BindingOperation's input, if one
	 * 					could be found (null or "" if not)
	 */
	public String getMethodInputNamespace (BindingOperation bindOp)
	{
		// Okay, this is an unbelievably cheesy way to get the Input namespace,
		// but I just didn't know another way to do it. What you should be able to
		// do is get the BindingInput as a QName, and extract the namespace from
		// there, but I didn't know a way to do that. There's probably a really easy way,
		// but I didn't know it, so...
		String namespace = "";
		try {
			StringTokenizer st = new StringTokenizer(bindOp.getBindingInput().toString(), "\n");
			while (st.hasMoreTokens()) {
				String token = st.nextToken();
				if (token.toLowerCase().startsWith("namespaceuri=")) {
					namespace = token.substring(13);
					break;
				}
			}
		} catch (Exception ignored) {
		}
		
		// if we didn't find anything, or we returned an error, try just getting the namespace
		// of the binding operation
		if (namespace.length() == 0) {
			namespace = bindOp.getOperation().getInput().getMessage().getQName().getNamespaceURI();
		}
		
		return namespace;
	}
	
	
	/**
	 * For a given Port specification in a WSDL file, return the
	 * URL that should be used to make calls to this Port.
	 *
	 * @param port		the Port we're looking at
	 * @return		a String indicating the URL associated
	 * 					with this Port (or null if one could not be
	 * 					found)
	 */
	public String getPortSoapURL (Port port)
	{
		List els = port.getExtensibilityElements();
		for (int k = 0; k < els.size(); k++)
		{
			if (els.get(k) instanceof SOAPAddress)
				return ((SOAPAddress)els.get(k)).getLocationURI();
		}
		
		return null;
	}
	
	
	/**
	 * Add a complex data type to the SOAPMappingRegistry
	 * for any calls you make with this SoapHelper. This is
	 * necessary if complex data types are used as parameters
	 * or responses.
	 *
	 * @param namespace		the namespace this type belongs to.
	 * 							You can also pass an empty String
	 * 							if you don't want a namespace
	 * 							associated with this data type 
	 * 							(which could be the case if you're
	 * 							using it to deserialize a return
	 * 							value from a system that doesn't
	 * 							include data types with the return
	 * 							values -- often the case with Microsoft
	 * 							web services)
	 * @param datatypeName	the name of the complex data
	 * 							type you're mapping (i.e. -- the
	 * 							name specified in the WSDL file)
	 * @param datatypeClass	the Java Class you want to map
	 * 							to this complex data type, which
	 * 							should (obviously) have the same
	 * 							structure as the data type (see
	 * 							the Apache SOAP samples for
	 * 							examples of how this works). It
	 * 							also MUST have the characteristics
	 * 							of a JavaBean, with get/set statements
	 * 							for all variables that are in the complex
	 * 							data type you're modeling.
	 */
	public void addComplexDatatype (String namespace, 
		String datatypeName, Class datatypeClass)
		throws WSDLException
	{
		if (smr == null) {
			smr = new SOAPMappingRegistry();
		}
		
		BeanSerializer beanSer = new BeanSerializer();
		smr.mapTypes(org.apache.soap.Constants.NS_URI_SOAP_ENC,
				new org.apache.soap.util.xml.QName(namespace, datatypeName),
				datatypeClass, beanSer, beanSer);
		debugPrint("Added mapping for " + datatypeName + " (" + namespace + ")");
	}
	
	
	/**
	 * Same as addComplexDatatype(namespace, datatypeName, datatypeClass),
	 * only the namespace is determined by searching through the types
	 * that are returned by the getTypes method. If the datatypeName
	 * is not found in the getTypes list, "" is used for the namespace.
	 * 
	 * @param datatypeName	the name of the complex data
	 * 							type you're mapping (i.e. -- the
	 * 							name specified in the WSDL file)
	 * @param datatypeClass	the Java Class you want to map
	 * 							to this complex data type
	 * @throws WSDLException
	 */
	public void addComplexDatatype (String datatypeName, 
		Class datatypeClass)
		throws WSDLException
	{
		addComplexDatatype(getNamespaceForType(datatypeName, caseSensitive), 
				datatypeName, datatypeClass);
	}
	
	
	/**
	 * Attempt to find and return the namespace used by the given
	 * complex data type. The caseSensitive parameter determines
	 * whether the data type name match is case-sensitive or not,
	 * as we're searching the type list in the WSDL file.
	 * <p> 
	 * The type list is generated by looking at the input and output
	 * message definitions for each method, so it's possible that 
	 * there are complex data types that are defined in the WSDL file 
	 * that won't be listed here, because they aren't used as 
	 * parameters or return values. So, for example, a complex data
	 * type that is nested within another complex data type might
	 * not show up on this list. 
	 *
	 * @param typeName	the name of the data type we're looking for
	 * @param caseSensitive		whether or not the name match
	 * 							should be case-sensitive
	 * @return		a String representing the namespace used by
	 * 						this data type, or "" if the type name
	 * 						was not found 
	 * @throws WSDLException
	 */
	public String getNamespaceForType (String typeName, boolean caseSensitive)
		throws WSDLException
	{
		Map typeMap = getTypes();
		try {
			if (caseSensitive) {
				return (String)typeMap.get(typeName);
			} else {
				for (Iterator it = typeMap.keySet().iterator(); it.hasNext(); ) {
					String tName = (String)it.next();
					if (tName.equalsIgnoreCase(typeName)) {
						return (String)typeMap.get(tName);
					}
				}
			}
		} catch (Exception e) {
			// TODO: handle exception
		}
		
		return "";
	}
	
	
	/**
	 * Add a data type to the SOAPMappingRegistry for values that
	 * are returned by calls you make with this SoapHelper. This is
	 * necessary the web service you are calling does not include
	 * data types with the XML elements in the SOAP body that
	 * is returned (which is common with Microsoft web services).
	 * <p>
	 * Normally, Apache SOAP will automatically deserialize common
	 * data types based on the type that is given for each value (like 
	 * xsd:string, xsd:int, etc.). But if the return values don't have any
	 * data types specified, you'll have to provide mappings for each
	 * value that is returned. For example:
	 * <p><blockquote><pre>
	 * wsw.addReturnDatatype("Result", new StringDeserializer());
	 * </pre></blockquote><p>
	 * Another use of this method is to specify that a response parameter
	 * should be returned as a raw XML node, with no translation, using 
	 * the XmlNodeDeserializer class like this:
	 * <p><blockquote><pre>
	 * wsw.addReturnDatatype("Result", new XmlNodeDeserializer());
	 * </pre></blockquote><p>
	 * This method attempts to guess what the namespace of the type
	 * will be, based on the types that are used as input and output
	 * parameters of the methods defined in the WSDL file. If you want
	 * to use an empty namespace instead, use the version of this
	 * method that allows you to specify a namespace, and pass "" as
	 * the namespace parameter.
	 *
	 * @param datatypeName	the name of the data type you're 
	 * 								mapping (i.e. -- the name of the parameter
	 * 								that is being returned, like "Result")
	 * @param deserializer		a Deserializer that will be used to read
	 * 								this data. A number of deserializers for
	 * 								common data types are available in 
	 * 								org.apache.soap.encoding.soapenc.* 
	 * 								(StringDeserializer, ArrayDeserializer, etc)
	 */
	public void addReturnDatatype (String datatypeName, Deserializer deserializer)
	throws WSDLException
	{
		addReturnDatatype(datatypeName, getNamespaceForType(datatypeName, false), 
				deserializer);
	}
	
	
	/**
	 * This version of the addReturnDatatype method allows you to
	 * specify a particular namespace. Otherwise, see the description
	 * of the other version of this method for more information about
	 * what it does.
	 * 
	 * @param datatypeName
	 * @param namespace
	 * @param deserializer
	 * @throws WSDLException
	 */
	public void addReturnDatatype (String datatypeName, String namespace, Deserializer deserializer)
		throws WSDLException
	{
		if (smr == null) {
			smr = new SOAPMappingRegistry();
		}
		
		smr.mapTypes(org.apache.soap.Constants.NS_URI_SOAP_ENC, 
			new org.apache.soap.util.xml.QName(namespace, datatypeName), null, null, 
			deserializer);
		debugPrint("Added mapping for " + datatypeName + " (" + deserializer.getClass().getName() + ")");
	}
	
	
	/**
	 * For a given Envelope, Mapping, and Context, return the
	 * SOAP message it represents, in raw XML form.
	 *
	 * @param env		an Envelope object, normally from a
	 * 							call to Call.buildEnvelope() or
	 * 							Response.buildEnvelope()
	 * @param smr		a SOAPMappingRegistry, normally from
	 * 							Call.getSOAPMappingRegistry()
	 * @param ctx		a SOAPContext, normally from 
	 * 							RPCMessage.getSOAPContext()
	 * @return		a String representing the XML SOAP message
	 * 					that would be generated from these parameters,
	 * 					or an error message if there was a problem
	 */
	public String getRawSOAPMessage (Envelope env, SOAPMappingRegistry smr, SOAPContext ctx)
	{
		try {
			java.io.StringWriter sw = new java.io.StringWriter();
			env.marshall(sw, smr, ctx);
			return sw.toString();
		} catch (Exception e) {
			return e.getMessage();
		}
	}
	
	/**
	 * For the given SOAP Call object, return the
	 * SOAP message it represents, in raw XML form.
	 *
	 * @param call		a Call object that's ready to be sent
	 * @return		a String representing the XML SOAP message
	 * 					that would be generated by this Call object,
	 * 					or an error message if there was a problem
	 */
	public String getRawSOAPMessage (Call call)
	{
		return getRawSOAPMessage(call.buildEnvelope(), 
				call.getSOAPMappingRegistry(), 
				call.getSOAPContext());
	}
	
	
	/**
	 * For the given SOAP Call and Response, return the
	 * SOAP message that the Response represents, in 
	 * raw XML form (the Call should be the Call that produced
	 * the Response, which is necessary to obtain the
	 * SOAPRegistryMapping that was used).
	 * <p>
	 * Please note that the XML string that is generated by 
	 * this version of the getRawSOAPMessage method is 
	 * NOT necessarily the same as the actual message that 
	 * was received. It is simply a represenation of the information
	 * that was parsed into the given Response object. If you want
	 * the actual message that was returned you will need to use
	 * a sniffer or proxy like TCPMon.
	 *
	 * @param resp		a Response object that has been received
	 * 							as a SOAP response to a Call
	 * @param call		a Call object that has been sent
	 * @return		a String representing the XML SOAP message
	 * 					that could have created this Response object,
	 * 					or an error message if there was a problem
	 */
	public String getRawSOAPResponse (Response resp, Call call)
	{
		return getRawSOAPMessage(resp.buildEnvelope(), 
				call.getSOAPMappingRegistry(), 
				resp.getSOAPContext());
	}
	
	
}

